package jaybot;
import java.util.ArrayList;
import java.util.LinkedList;

import aiinterface.AIInterface;
import aiinterface.CommandCenter;
import enumerate.Action;
import enumerate.State;
import simulator.Simulator;
import struct.CharacterData;
import struct.FrameData;
import struct.GameData;
import struct.Key;
import struct.MotionData;

/**
 * BASIC MCTS AI
 *
 * @author Taichi
 *
 */

/**
 * JayBot_2017 ( Using MCTS & Robustic Action-Table AI)
 *
 * @author Man-Je Kim
 *
 */
public class JayBot_2017 implements AIInterface {

	private Simulator simulator;
	private Key key;
	private CommandCenter commandCenter;
	private boolean playerNumber;
	private GameData gameData;

	/** 大本?FrameData */
	private FrameData frameData;

	/** 大本よりFRAME_AHEAD分?れ?FrameData */
	private FrameData simulatorAheadFrameData;

	/** 自分?行?る行動全? */
	private LinkedList<Action> myActions;

	/** 相手?行?る行動全? */
	private LinkedList<Action> oppActions;

	/** 自分?情報 */
	private CharacterData myCharacter;

	/** 相手?情報 */
	private CharacterData oppCharacter;

	/** フレーム?調整用時間(JerryMizunoAIを?考) */
	private static final int FRAME_AHEAD = 14;

	private ArrayList<MotionData> myMotion;

	private ArrayList<MotionData> oppMotion;

	private Action[] actionAir;

	private Action[] actionGround;

	private Action spSkill;

	private Node rootNode;

	private CharacterName charName;

	/** デ?ッグモード??る????。true?場??様々?ログ?出力?れる */
	public static final boolean DEBUG_MODE = false;

	@Override
	public void close() {

	}

	@Override
	public void getInformation(FrameData frameData) {
		this.frameData = frameData;
		this.commandCenter.setFrameData(this.frameData, playerNumber);

		myCharacter = frameData.getCharacter(playerNumber);
		oppCharacter = frameData.getCharacter(!playerNumber);
	}

	@Override
	public int initialize(GameData gameData, boolean playerNumber) {
		this.playerNumber = playerNumber;
		this.gameData = gameData;

		this.key = new Key();
		this.frameData = new FrameData();
		this.commandCenter = new CommandCenter();

		this.myActions = new LinkedList<Action>();
		this.oppActions = new LinkedList<Action>();

		simulator = gameData.getSimulator();
		actionAir = new Action[] { Action.AIR_GUARD, Action.AIR_A, Action.AIR_B, Action.AIR_DA, Action.AIR_DB,
				Action.AIR_FA, Action.AIR_FB, Action.AIR_UA, Action.AIR_UB, Action.AIR_D_DF_FA, Action.AIR_D_DF_FB,
				Action.AIR_F_D_DFA, Action.AIR_F_D_DFB, Action.AIR_D_DB_BA, Action.AIR_D_DB_BB };
		actionGround = new Action[] { Action.STAND_D_DB_BA, Action.BACK_STEP, Action.FORWARD_WALK, Action.DASH,
				Action.JUMP, Action.FOR_JUMP, Action.BACK_JUMP, Action.STAND_GUARD, Action.CROUCH_GUARD, Action.THROW_A,
				Action.THROW_B, Action.STAND_A, Action.STAND_B, Action.CROUCH_A, Action.CROUCH_B, Action.STAND_FA,
				Action.STAND_FB, Action.CROUCH_FA, Action.CROUCH_FB, Action.STAND_D_DF_FA, Action.STAND_D_DF_FB,
				Action.STAND_F_D_DFA, Action.STAND_F_D_DFB, Action.STAND_D_DB_BB };
		spSkill = Action.STAND_D_DF_FC;

		myMotion = gameData.getMotionData(playerNumber);
		oppMotion = gameData.getMotionData(!playerNumber);

		String tmpcharname = this.gameData.getCharacterName(this.playerNumber);
		if (tmpcharname.equals("ZEN"))
			charName = CharacterName.ZEN;
		else if (tmpcharname.equals("GARNET"))
			charName = CharacterName.GARNET;
		else if (tmpcharname.equals("LUD"))
			charName = CharacterName.LUD;
		else
			charName = CharacterName.OTHER;

		return 0;
	}

	@Override
	public Key input() {
		return key;
	}

	@Override
	public void processing() {

		if (canProcessing()) {
			if (commandCenter.getSkillFlag()) {
				key = commandCenter.getSkillKey();
			} else {
				key.empty();
				commandCenter.skillCancel();

				mctsPrepare(); // MCTS?下準備を行?
				rootNode = new Node(simulatorAheadFrameData, null, myActions, oppActions, gameData, playerNumber,
						commandCenter);
				rootNode.createNode();

				Action bestAction = rootNode.mcts(); // MCTS?実行
				if (JayBot_2017.DEBUG_MODE) {
					rootNode.printNode(rootNode);
				}

				commandCenter.commandCall(bestAction.name()); // MCTS??択?れ?行動を実行?る
			}
		}
	}

	/**
	 * AI?行動???る????を判別?る
	 *
	 * @return AI?行動???る????
	 */
	public boolean canProcessing() {
		return !frameData.getEmptyFlag() && frameData.getRemainingTime() > 0;
	}

	/**
	 * MCTS?下準備 <br>
	 * 14フレーム進???FrameData??得??を行?
	 */
	public void mctsPrepare() {
		simulatorAheadFrameData = simulator.simulate(frameData, playerNumber, null, null, FRAME_AHEAD);

		myCharacter = simulatorAheadFrameData.getCharacter(playerNumber);
		oppCharacter = simulatorAheadFrameData.getCharacter(!playerNumber);

		setMyAction();
		setOppAction();
	}

	public void setMyAction() {
		myActions.clear();

		int energy = myCharacter.getEnergy();

		if (charName.name() == "ZEN") {
			if (myCharacter.getState() == State.AIR) {
				for (int i = 0; i < actionAir.length; i++) {
					if (Math.abs(myMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
							.getAttackStartAddEnergy()) <= energy) {
						myActions.add(actionAir[i]);
					}
				}
			} else if (simulatorAheadFrameData.getDistanceX() < 50) {

				myActions.add(Action.STAND_D_DF_FA);
				myActions.add(Action.STAND_B);
				myActions.add(Action.CROUCH_B);
				myActions.add(Action.STAND_D_DB_BA);
				if (myCharacter.getEnergy() >= 50)
					myActions.add(Action.STAND_D_DB_BB);
			}

			else if (simulatorAheadFrameData.getDistanceX() < 85) {

				myActions.add(Action.STAND_B);
				myActions.add(Action.AIR_B);
				myActions.add(Action.CROUCH_B);
				myActions.add(Action.AIR_DB);
				if (myCharacter.getEnergy() >= 50)
					myActions.add(Action.STAND_D_DB_BB);
			} else if (simulatorAheadFrameData.getDistanceX() < 100) {

				myActions.add(Action.STAND_B);
				myActions.add(Action.CROUCH_FB);
				myActions.add(Action.CROUCH_B);
				myActions.add(Action.STAND_FB);
				myActions.add(Action.AIR_DB);
				if (myCharacter.getEnergy() >= 50)
					myActions.add(Action.STAND_D_DB_BB);
			} else if (simulatorAheadFrameData.getDistanceX() < 105) {
				myActions.add(Action.CROUCH_FB);
				myActions.add(Action.CROUCH_B);
				myActions.add(Action.STAND_FB);
				myActions.add(Action.STAND_D_DF_FA);
				myActions.add(Action.AIR_DB);
				if (myCharacter.getEnergy() >= 30)
					myActions.add(Action.STAND_D_DF_FB);
				if (myCharacter.getEnergy() >= 50)
					myActions.add(Action.STAND_D_DB_BB);
			} else {
				if (Math.abs(
						myMotion.get(Action.valueOf(spSkill.name()).ordinal()).getAttackStartAddEnergy()) <= energy) {
					myActions.add(spSkill);
				}

				for (int i = 0; i < actionGround.length; i++) {
					if (Math.abs(myMotion.get(Action.valueOf(actionGround[i].name()).ordinal())
							.getAttackStartAddEnergy()) <= energy) {
						myActions.add(actionGround[i]);
					}
				}
			}
		} else if (charName.name() == "LUD") {
			if (myCharacter.getState() == State.AIR) {
				for (int i = 0; i < actionAir.length; i++) {
					if (Math.abs(myMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
							.getAttackStartAddEnergy()) <= energy) {
						myActions.add(actionAir[i]);
					}
				}
			} else if (simulatorAheadFrameData.getDistanceX() < 50) {
				myActions.add(Action.AIR_DB);
				myActions.add(Action.AIR_B);
				myActions.add(Action.CROUCH_A);
				myActions.add(Action.CROUCH_B);
				myActions.add(Action.AIR_A);
				myActions.add(Action.CROUCH_FA);
				myActions.add(Action.STAND_B);
				if (myCharacter.getEnergy() >= 1)
					myActions.add(Action.STAND_A);
				if (myCharacter.getEnergy() >= 8)
					myActions.add(Action.STAND_FA);
				if (myCharacter.getEnergy() >= 10)
					myActions.add(Action.THROW_A);
				if (myCharacter.getEnergy() >= 20)
					myActions.add(Action.STAND_D_DB_BB);

			}

			else if (simulatorAheadFrameData.getDistanceX() < 85) {

				myActions.add(Action.STAND_B);
				myActions.add(Action.CROUCH_B);
				myActions.add(Action.AIR_B);
				myActions.add(Action.CROUCH_A);
				myActions.add(Action.CROUCH_FA);
				if (myCharacter.getEnergy() >= 30)
					myActions.add(Action.STAND_D_DB_BA);

				if (myCharacter.getEnergy() >= 30)
					myActions.add(Action.STAND_D_DF_FA);
				if (myCharacter.getEnergy() >= 20)
					myActions.add(Action.STAND_D_DF_FB);
				if (myCharacter.getEnergy() >= 10)
					myActions.add(Action.AIR_UB);
				if (myCharacter.getEnergy() >= 20)
					myActions.add(Action.STAND_FB);
				if (myCharacter.getEnergy() >= 30)
					myActions.add(Action.STAND_F_D_DFA);

			} else if (simulatorAheadFrameData.getDistanceX() < 100) {

				myActions.add(Action.CROUCH_B);
				myActions.add(Action.CROUCH_FA);
				myActions.add(Action.STAND_B);
				myActions.add(Action.CROUCH_A);
				myActions.add(Action.AIR_B);
				if (myCharacter.getEnergy() >= 8)
					myActions.add(Action.STAND_FA);
				if (myCharacter.getEnergy() >= 1)
					myActions.add(Action.STAND_A);
				if (myCharacter.getEnergy() >= 10)
					myActions.add(Action.THROW_A);
				if (myCharacter.getEnergy() >= 20)
					myActions.add(Action.CROUCH_FB);
				if (myCharacter.getEnergy() >= 30)
					myActions.add(Action.STAND_F_D_DFA);
			} else {
				if (Math.abs(
						myMotion.get(Action.valueOf(spSkill.name()).ordinal()).getAttackStartAddEnergy()) <= energy) {
					myActions.add(spSkill);
				}

				for (int i = 0; i < actionGround.length; i++) {
					if (Math.abs(myMotion.get(Action.valueOf(actionGround[i].name()).ordinal())
							.getAttackStartAddEnergy()) <= energy) {
						myActions.add(actionGround[i]);
					}
				}
			}
		}

		else if (charName.name() == "GARNET") {
			if (myCharacter.getState() == State.AIR) {
				for (int i = 0; i < actionAir.length; i++) {
					if (Math.abs(myMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
							.getAttackStartAddEnergy()) <= energy) {
						myActions.add(actionAir[i]);
					}
				}
			} else if (simulatorAheadFrameData.getDistanceX() < 50) {

				myActions.add(Action.STAND_D_DB_BA);
				myActions.add(Action.CROUCH_B);
				myActions.add(Action.STAND_F_D_DFA);
				myActions.add(Action.STAND_A);
				myActions.add(Action.STAND_FB);
				if (myCharacter.getEnergy() >= 50)
					myActions.add(Action.STAND_F_D_DFB);
				if (myCharacter.getEnergy() >= 30)
					myActions.add(Action.STAND_D_DF_FA);
				if (myCharacter.getEnergy() >= 25)
					myActions.add(Action.STAND_D_DF_FB);
			}

			else if (simulatorAheadFrameData.getDistanceX() < 85) {

				myActions.add(Action.STAND_FB);
				myActions.add(Action.AIR_UB);
				myActions.add(Action.STAND_D_DB_BA);
				myActions.add(Action.STAND_B);
				myActions.add(Action.CROUCH_B);
				if (myCharacter.getEnergy() >= 25)
					myActions.add(Action.STAND_D_DF_FB);
				if (myCharacter.getEnergy() >= 30)
					myActions.add(Action.STAND_D_DF_FA);
			} else if (simulatorAheadFrameData.getDistanceX() < 100) {

				myActions.add(Action.STAND_FB);
				myActions.add(Action.STAND_D_DB_BA);
				myActions.add(Action.STAND_F_D_DFA);
				myActions.add(Action.STAND_B);
				myActions.add(Action.CROUCH_B);
				if (myCharacter.getEnergy() >= 25)
					myActions.add(Action.STAND_D_DF_FB);
				if (myCharacter.getEnergy() >= 30)
					myActions.add(Action.STAND_D_DF_FA);

			}

			else {
				if (Math.abs(
						myMotion.get(Action.valueOf(spSkill.name()).ordinal()).getAttackStartAddEnergy()) <= energy) {
					myActions.add(spSkill);
				}

				for (int i = 0; i < actionGround.length; i++) {
					if (Math.abs(myMotion.get(Action.valueOf(actionGround[i].name()).ordinal())
							.getAttackStartAddEnergy()) <= energy) {
						myActions.add(actionGround[i]);
					}
				}
			}
		}
	}

	public void setOppAction() {
		oppActions.clear();

		int energy = oppCharacter.getEnergy();

		if (charName.name() == "ZEN") {
			if (oppCharacter.getState() == State.AIR) {
				for (int i = 0; i < actionAir.length; i++) {
					if (Math.abs(oppMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
							.getAttackStartAddEnergy()) <= energy) {
						oppActions.add(actionAir[i]);
					}
				}
			} else if (simulatorAheadFrameData.getDistanceX() < 50) {

				oppActions.add(Action.CROUCH_FA);
				oppActions.add(Action.STAND_B);
				oppActions.add(Action.CROUCH_B);
				oppActions.add(Action.AIR_B);
				oppActions.add(Action.AIR_DB);
				if (oppCharacter.getEnergy() >= 30)
					oppActions.add(Action.STAND_D_DF_FB);
				if (oppCharacter.getEnergy() >= 50)
					oppActions.add(Action.STAND_D_DB_BB);
			}

			else if (simulatorAheadFrameData.getDistanceX() < 85) {

				oppActions.add(Action.STAND_B);
				oppActions.add(Action.CROUCH_FA);
				oppActions.add(Action.AIR_B);
				oppActions.add(Action.CROUCH_B);
				oppActions.add(Action.AIR_DB);
				if (oppCharacter.getEnergy() >= 30)
					oppActions.add(Action.STAND_D_DF_FB);
				if (oppCharacter.getEnergy() >= 50)
					oppActions.add(Action.STAND_D_DB_BB);
			} else if (simulatorAheadFrameData.getDistanceX() < 100) {

				oppActions.add(Action.STAND_B);
				oppActions.add(Action.CROUCH_FB);
				oppActions.add(Action.CROUCH_B);
				oppActions.add(Action.STAND_FB);
				oppActions.add(Action.AIR_DB);
				if (oppCharacter.getEnergy() >= 30)
					oppActions.add(Action.STAND_D_DF_FB);
				if (oppCharacter.getEnergy() >= 50)
					oppActions.add(Action.STAND_D_DB_BB);
			} else if (simulatorAheadFrameData.getDistanceX() < 105) {
				oppActions.add(Action.CROUCH_FB);
				oppActions.add(Action.CROUCH_B);
				oppActions.add(Action.STAND_FB);
				oppActions.add(Action.STAND_D_DF_FA);
				oppActions.add(Action.AIR_DB);
				if (oppCharacter.getEnergy() >= 30)
					oppActions.add(Action.STAND_D_DF_FB);
				if (oppCharacter.getEnergy() >= 50)
					oppActions.add(Action.STAND_D_DB_BB);
			}

			else {
				if (Math.abs(
						oppMotion.get(Action.valueOf(spSkill.name()).ordinal()).getAttackStartAddEnergy()) <= energy) {
					oppActions.add(spSkill);
				}

				for (int i = 0; i < actionGround.length; i++) {
					if (Math.abs(oppMotion.get(Action.valueOf(actionGround[i].name()).ordinal())
							.getAttackStartAddEnergy()) <= energy) {
						oppActions.add(actionGround[i]);
					}
				}
			}
		} else if (charName.name() == "LUD") {
			if (oppCharacter.getState() == State.AIR) {
				for (int i = 0; i < actionAir.length; i++) {
					if (Math.abs(oppMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
							.getAttackStartAddEnergy()) <= energy) {
						oppActions.add(actionAir[i]);
					}
				}
			} else if (simulatorAheadFrameData.getDistanceX() < 50) {
				oppActions.add(Action.AIR_DB);
				oppActions.add(Action.AIR_B);
				oppActions.add(Action.CROUCH_A);
				oppActions.add(Action.CROUCH_B);
				oppActions.add(Action.AIR_A);
				oppActions.add(Action.CROUCH_FA);
				oppActions.add(Action.STAND_B);
				if (oppCharacter.getEnergy() >= 1)
					oppActions.add(Action.STAND_A);
				if (oppCharacter.getEnergy() >= 8)
					oppActions.add(Action.STAND_FA);
				if (oppCharacter.getEnergy() >= 10)
					oppActions.add(Action.THROW_A);
				if (oppCharacter.getEnergy() >= 20)
					oppActions.add(Action.STAND_D_DB_BB);

			}

			else if (simulatorAheadFrameData.getDistanceX() < 85) {

				oppActions.add(Action.STAND_B);
				oppActions.add(Action.CROUCH_B);
				oppActions.add(Action.AIR_B);
				oppActions.add(Action.CROUCH_A);
				oppActions.add(Action.CROUCH_FA);
				if (oppCharacter.getEnergy() >= 30)
					oppActions.add(Action.STAND_D_DB_BA);

				if (oppCharacter.getEnergy() >= 30)
					oppActions.add(Action.STAND_D_DF_FA);
				if (oppCharacter.getEnergy() >= 20)
					oppActions.add(Action.STAND_D_DF_FB);
				if (oppCharacter.getEnergy() >= 10)
					oppActions.add(Action.AIR_UB);
				if (oppCharacter.getEnergy() >= 20)
					oppActions.add(Action.STAND_FB);
				if (oppCharacter.getEnergy() >= 30)
					oppActions.add(Action.STAND_F_D_DFA);

			} else if (simulatorAheadFrameData.getDistanceX() < 100) {

				oppActions.add(Action.CROUCH_B);
				oppActions.add(Action.CROUCH_FA);
				oppActions.add(Action.STAND_B);
				oppActions.add(Action.CROUCH_A);
				oppActions.add(Action.AIR_B);
				if (oppCharacter.getEnergy() >= 8)
					oppActions.add(Action.STAND_FA);
				if (oppCharacter.getEnergy() >= 1)
					oppActions.add(Action.STAND_A);
				if (oppCharacter.getEnergy() >= 10)
					oppActions.add(Action.THROW_A);
				if (oppCharacter.getEnergy() >= 20)
					oppActions.add(Action.CROUCH_FB);
				if (oppCharacter.getEnergy() >= 30)
					oppActions.add(Action.STAND_F_D_DFA);
			} else {
				if (Math.abs(
						oppMotion.get(Action.valueOf(spSkill.name()).ordinal()).getAttackStartAddEnergy()) <= energy) {
					oppActions.add(spSkill);
				}

				for (int i = 0; i < actionGround.length; i++) {
					if (Math.abs(oppMotion.get(Action.valueOf(actionGround[i].name()).ordinal())
							.getAttackStartAddEnergy()) <= energy) {
						oppActions.add(actionGround[i]);
					}
				}
			}
		} else if (charName.name() == "GARNET") {
			if (oppCharacter.getState() == State.AIR) {
				for (int i = 0; i < actionAir.length; i++) {
					if (Math.abs(oppMotion.get(Action.valueOf(actionAir[i].name()).ordinal())
							.getAttackStartAddEnergy()) <= energy) {
						oppActions.add(actionAir[i]);
					}
				}
			} else if (simulatorAheadFrameData.getDistanceX() < 50) {

				oppActions.add(Action.STAND_D_DB_BA);
				oppActions.add(Action.CROUCH_B);
				oppActions.add(Action.STAND_F_D_DFA);
				oppActions.add(Action.STAND_A);
				oppActions.add(Action.STAND_FB);
				if (oppCharacter.getEnergy() >= 50)
					oppActions.add(Action.STAND_F_D_DFB);
				if (oppCharacter.getEnergy() >= 30)
					oppActions.add(Action.STAND_D_DF_FA);
				if (oppCharacter.getEnergy() >= 25)
					oppActions.add(Action.STAND_D_DF_FB);
			}

			else if (simulatorAheadFrameData.getDistanceX() < 85) {

				oppActions.add(Action.STAND_FB);
				oppActions.add(Action.AIR_UB);
				oppActions.add(Action.STAND_D_DB_BA);
				oppActions.add(Action.STAND_B);
				oppActions.add(Action.CROUCH_B);
				if (oppCharacter.getEnergy() >= 25)
					oppActions.add(Action.STAND_D_DF_FB);
				if (oppCharacter.getEnergy() >= 30)
					oppActions.add(Action.STAND_D_DF_FA);
			} else if (simulatorAheadFrameData.getDistanceX() < 100) {

				oppActions.add(Action.STAND_FB);
				oppActions.add(Action.STAND_D_DB_BA);
				oppActions.add(Action.STAND_F_D_DFA);
				oppActions.add(Action.STAND_B);
				oppActions.add(Action.CROUCH_B);
				if (oppCharacter.getEnergy() >= 25)
					oppActions.add(Action.STAND_D_DF_FB);
				if (oppCharacter.getEnergy() >= 30)
					oppActions.add(Action.STAND_D_DF_FA);

			}

			else {
				if (Math.abs(
						oppMotion.get(Action.valueOf(spSkill.name()).ordinal()).getAttackStartAddEnergy()) <= energy) {
					oppActions.add(spSkill);
				}

				for (int i = 0; i < actionGround.length; i++) {
					if (Math.abs(oppMotion.get(Action.valueOf(actionGround[i].name()).ordinal())
							.getAttackStartAddEnergy()) <= energy) {
						oppActions.add(actionGround[i]);
					}
				}
			}
		}
	}

	public boolean EqulCheck(Action a, Action b) {

		if (a == b)
			return true;
		else
			return false;
	}

	public enum CharacterName {
		ZEN, GARNET, LUD, OTHER;

	}

	@Override
	public void roundEnd(int arg0, int arg1, int arg2) {
		commandCenter.skillCancel();
		key.empty();
	}

}
